# Undefined Chromium oxide crystals used as calibrant on some MX beamlines at ESRF.
# d-spacing measures on the Powder diffraction beamline, formerly ID31.  
# Ref: Personal communication from scientist, do not use. 
3.645
2.672
2.487
2.181
1.819
1.676
1.467
1.433
